/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2015
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2015
 * Copyright (C) RWTH Aachen University, Federal Republic of Germany, 2013-2015
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CProtSharedMemory.h
 *       A shared memory implementation of the communication protocol interface.
 *
 *  This implementation may be used to communicate between processes that share
 *  memory.
 *
 * @author Joachim Protze
 * @date 15.03.2012
 * 
 */

#ifndef C_PROT_SHARED_MEMORY_H
#define C_PROT_SHARED_MEMORY_H

#include "I_CommProtocol.h"
#include "I_Module.h"
#include "ModuleBase.h"

#include <vector>
#include <list>
#include <map>
#include <sys/types.h>

#define GTI_SHM_SHORT_MSG_SIZE 4050

namespace gti
{

    typedef int pipe_t;

    typedef enum{
        GTI_SHM_SHORT_MSG=0,
        GTI_SHM_LONG_MSG
    }gti_shm_msg_type;

    enum gti_shm_err_type{
        MSGGET_ERROR=-100,
        MSGSND_ERROR,
        MSGRCV_ERROR,
        MSGCTL_ERROR,
        SHMGET_ERROR,
        SHMAT_ERROR,
        SHMDT_ERROR,
        SHMCTL_ERROR,
        INTERNAL_ERROR,
    // highest shm_err_type should be -2
        LAST_SHM_ERROR // < 0 by assert
    };


    struct msgbuf {
        long             mtype;         /* message type, must be > 0 */
        ssize_t          length;        /* length of mtext */
        gti_shm_msg_type type;          /* message or shmem */
        char             mtext[GTI_SHM_SHORT_MSG_SIZE];   /* message data */
    };

    struct msgpointer {
        long             mtype;         /* message type, must be > 0 */
        ssize_t          length;        /* size of shm-block */
        gti_shm_msg_type type;          /* message or shmem */
        pipe_t           shmid;           /* message data */
    };

    struct shmRequest{
        void* out_buf;
        unsigned long num_bytes;
        unsigned int request;
        ssize_t length;
        unsigned long channel;
        int finished;
        bool send;
        shmRequest(unsigned int id, void* out_buf, unsigned long num_bytes, unsigned long channel, bool send=false) : request(id), finished(false), out_buf(out_buf), num_bytes(num_bytes), channel(channel), send(send) {}
        shmRequest(const shmRequest& o) : out_buf(o.out_buf), num_bytes(o.num_bytes), request(o.request), length(o.length), channel(o.channel), finished(o.finished), send(o.send) {}
    };


    /**
     * Class that describes the communication protocol interface.
     */
    class CommProtSharedMemory : public ModuleBase<CommProtSharedMemory, I_CommProtocol>
    {
    protected:
        bool initialized;
        bool finalized;
        bool isTop;
        bool myIsIntra;
        
        std::vector<pipe_t>   localPipes;
        std::vector<key_t>  localKeys;

        std::vector<pipe_t> remotePipes;
        std::vector<key_t> remoteKeys;
        std::vector<key_t>   openLongSendKeys;
        
        int numPartners;
        int mySeed;

        std::map<key_t, int> key2Channel;

        GTI_DISTRIBUTION distrib;
        int blocksize;
        unsigned long gtiOwnLevel;
        unsigned long remoteTierSize;
        unsigned long tierSize;
        unsigned long selfID;
        int commId;
        unsigned long myPlaceId;
        char commSide;

        std::map<int, shmRequest*> requestMap;
        unsigned int requestId;
        
        
        /**
         * initialize connections
         */
        void connect();
        ssize_t recv_wrapper ( void* out_buf, size_t num_bytes, unsigned long channel, unsigned long* out_channel, int msgflg=0 );
    public:

        /**
         * Constructor.
         * @ref ModConf - The module configuration syntax
         * @param intanceName name of the module instance.
         */
        CommProtSharedMemory (const char* instanceName);

        /**
         * Destructor.
         */
        ~CommProtSharedMemory (void);

        /**
         * @see gti::I_CommProtocol::isConnected
         */
        bool isConnected (void);

        /**
         * @see gti::I_CommProtocol::isInitialized
         */
        bool isInitialized (void);

        /**
         * @see gti::I_CommProtocol::isFinalized
         */
        bool isFinalized (void);

        /**
         * @see gti::I_CommProtocol::getNumChannels
         */
        GTI_RETURN getNumChannels (unsigned long* out_numChannels);

        /**
         * @see gti::I_CommProtocol::getPlaceId
         */
        GTI_RETURN getPlaceId (unsigned long* outPlaceId);

        /**
         * @see gti::I_CommProtocol::shutdown
         */
        GTI_RETURN shutdown (void);

        /**
         * @see gti::I_CommProtocol::removeOutstandingRequests
         */
        GTI_RETURN removeOutstandingRequests (void);

        /**
         * @see gti::I_CommProtocol::ssend
         */
        GTI_RETURN ssend (
                void* buf,
                unsigned long num_bytes,
                unsigned long channel
                );

        /**
         * @see gti::I_CommProtocol::isend
         */
        GTI_RETURN isend (
                void* buf,
                unsigned long num_bytes,
                unsigned int *out_request,
                unsigned long channel
                );

        /**
         * @see gti::I_CommProtocol::recv
         */
        GTI_RETURN recv (
                void* out_buf,
                unsigned long num_bytes,
                unsigned long* out_length,
                unsigned long channel,
                unsigned long *out_channel
                );

        /**
         * @see gti::I_CommProtocol::irecv
         */
        GTI_RETURN irecv (
                void* out_buf,
                unsigned long num_bytes,
                unsigned int *out_request,
                unsigned long channel
                );

        /**
         * @see gti::I_CommProtocol::test_msg
         */
        GTI_RETURN test_msg (
                unsigned int request,
                int* out_completed,
                unsigned long* out_receive_length,
                unsigned long* out_channel
                );

        /**
         * @see gti::I_CommProtocol::wait_msg
         */
        GTI_RETURN wait_msg (
                unsigned int request,
                unsigned long* out_receive_length,
                unsigned long* out_channel
                );
    }; /*class CommProtSharedMemory*/
} /*namespace gti*/

#endif /* C_PROT_SHARED_MEMORY_H */
