/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2015
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2015
 * Copyright (C) RWTH Aachen University, Federal Republic of Germany, 2013-2015
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratSimpleIntra.h
 *       @see I_CommStrategyIntra.
 *
 * @author Tobias Hilbrich
 * @date 16.01.2012
 */

#include "GtiDefines.h"
#include "GtiEnums.h"
#include "GtiTypes.h"
#include "I_CommStrategyIntra.h"
#include "I_CommProtocol.h"
#include "ModuleBase.h"
#include "CStratQueue.h"
#include "CStratIsend.h"

#ifndef CSTRAT_SIMPLE_INTRA_H
#define CSTRAT_SIMPLE_INTRA_H

namespace gti
{
    /**
     * Simple implementation of intra communication strategy.
     * Properties:
     * - no aggregation
     * - uses isend to send
     */
    class CStratIsendIntra : public ModuleBase<CStratIsendIntra, CStratIntraQueue>, public CStratIsend
    {
    protected:
        /** The communication protocol. */
        I_CommProtocol *protocol;

        //Receive  - irecv data
        unsigned int myRequest;
        unsigned long myBuf[2];

        ////Data to implement communicationFinished
        //Tokens
        static const unsigned long myTokenUpdate;
        static const unsigned long myTokenAcknowledge;

        //For everyone
        long myNumMsgsSent; //Number of sends we had
        long myNumMsgsReceived; //Number of successful tests or waits we had
        std::map <int, std::list<CStratQueueItem> > myReceivedUnexpectedMessages;

        //For Master
        long mySumDiffs; //Intermediate sum of received updates
        int myNumUpdates;
        std::vector<bool> myChannelState;

        bool myCommFinished;

        /**
         * Handles a length token and receives the described message.
         * @param tokenMessage the token message (2 unsigned longs).
         * @param channel from which we got the token message.
         * @param outNumBytes pointer to storage for actual message length.
         * @param outBuf pointer to void* that is set to a buffer that contains the message.
         * @param outFreeData data to use when we free the buffer.
         * @param outBufFreeFunction function to free the returned buffer.
         */
        GTI_RETURN handleReceivedMessageToken (
                unsigned long* tokenMessage,
                unsigned long channel,
                unsigned long* outNumBytes,
                void** outBuf,
                void** outFreeData,
                GTI_RETURN (**outBufFreeFunction) (void* freeData, unsigned long numBytes, void* buf)
                );

        /**
         * Checks whether an unexpected message exists that was received
         * during communicationFinished, if so it puts the first of these into
         * the given fields for a retrieved message. If not it does not touches
         * the fields (except outFlag) and returns false.
         *
         * For params see CstratIsendIntra::test
         * @return true if a unexpected message was available, false otherwise.
         */
        bool handleUnexpectedMessagesForReceive (
                int* outFlag,
                unsigned long *outFromPlace,
                unsigned long* outNumBytes,
                void** outBuf,
                void** outFreeData,
                GTI_RETURN (**outBufFreeFunction) (void* freeData, unsigned long numBytes, void* buf)
        );

        /**
         * Handles an unexpeced update, the master may receive
         * these.
         * @param buf token message of the update
         */
        void handleUnexpectedUpdate (unsigned long* buf, unsigned long channel);

        /**
         * Internal flush implementation, use block=true to block till completion.
         * @param block wait till completion?
         * @return GTI_SUCCESS iff successful.
         */
        GTI_RETURN flush (bool block);

        /**
         * Waits until the first send request in myRequests is completed.
         * It will not only wait for the communication but also check for
         * incoming messages to allow partners to continue their execution
         * if they also stall for a request completion.
         */
        void finishFirstSendRequest (void);

    public:

        /**
         * Constructor.
         * @ref ModConf see for details on module configuration.
         * @param instanceName name of the module instance.
         */
        CStratIsendIntra (const char* instanceName);

        /**
         *Destructor.
         */
        ~CStratIsendIntra (void);

        /**
         * @see I_CommStrategyIntra::shutdown
         */
        GTI_RETURN shutdown (GTI_FLUSH_TYPE flush_behavior);

        /**
         * @see I_CommStrategyIntra::communicationFinished
         */
        GTI_RETURN communicationFinished (bool *pOutIsFinished);

        /**
         * @see I_CommStrategyIntra::getNumPlaces
         */
        GTI_RETURN getNumPlaces (unsigned long *outNumPlaces);

        /**
         * @see I_CommStrategyIntra::getOwnPlaceId
         */
        GTI_RETURN getOwnPlaceId (unsigned long *outId);

        /**
         * @see I_CommStrategyIntra::send
         */
        GTI_RETURN send (
                unsigned long toPlace,
                void* buf,
                unsigned long numBytes,
                void* freeData,
                GTI_RETURN (*bufFreeFunction) (void* freeData, unsigned long numBytes, void* buf)
        );

        /**
         * @see I_CommStrategyIntra::test
         */
        GTI_RETURN test (
                int* outFlag,
                unsigned long *outFromPlace,
                unsigned long* outNumBytes,
                void** outBuf,
                void** outFreeData,
                GTI_RETURN (**outBufFreeFunction) (void* freeData, unsigned long numBytes, void* buf)
        );

        /**
         * @see I_CommStrategyIntra::wait
         */
        GTI_RETURN wait (
                unsigned long* outFromPlace,
                unsigned long* outNumBytes,
                void** outBuf,
                void** outFreeData,
                GTI_RETURN (**outBufFreeFunction) (void* freeData, unsigned long numBytes, void* buf)
        );

        /**
         * @see I_PanicListener::flush
         */
        GTI_RETURN flush (void);

        /**
         * @see I_PanicListener::flushAndSetImmediate
         */
        GTI_RETURN flushAndSetImmediate (void);
    };
}

#endif /* CSTRAT_SIMPLE_INTRA_H */
