/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C)
 *  2011-2015 ZIH, Technische Universitaet Dresden, Federal Republic of Germany
 *  2011-2015 Lawrence Livermore National Laboratories, United States of America
 *  2013-2015 RWTH Aachen University, Federal Republic of Germany
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file Mpize.cpp
 *  Calls MPI_Init and MPI_Finalize in applications that
 *  do not utilize MPI, to establish the basic GTI communication
 *  network.
 * 
 *
 * @author Felix Münchhalfen
 * @date 09.07.2014
 *
 */


#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <cstring>
#include "Mpize.h"

/**
 * The Mpize instance
 */
Mpize starter;

/**
 * This function gets called by PnMPI when the registration
 * points of all modules are processed.
 */
extern "C" void PNMPI_RegistrationComplete(void)
{
    starter.Init();
}

/**
 * Remembers if MPI_Init was already called
 */
int Mpize::initialized = 0;

/**
 * Constructor. Do not call MPI_Init here as
 * not all registration points have been passed
 * yet.
 */
Mpize::Mpize()
{

}

/**
 * Finalizes MPI when the module is unloaded.
 */
Mpize::~Mpize()
{
    if( initialized )
    {
        MPI_Finalize();
        initialized = 0;
    }
}

/**
 * Read command line arguments and initialize MPI
 */
void Mpize::Init()
{
    if( !initialized )
    {
        int argc;
        char **argv;
        read_cmdline(&argc, &argv);

        MPI_Init(&argc, &argv);
        initialized = 1;
    }
}

/**
 * Tries to demangle the argc and argv application
 * parameters out of /proc/self/cmdline or
 * in a OS specific way.
 * Parameters are seperated by \0 characters in this file.
 * 
 * TODO: Do these mechanism cover all relevant operating
 * systems?
 * 
 *
 * @param argc
 * @param argv
 */
#if defined(__linux__) || defined(linux) || defined(__linux)
void Mpize::read_cmdline( int* argc, char*** argv )
{
    FILE *fp = fopen("/proc/self/cmdline", "rb");
    if( fp == NULL )
    {
        printf("ERROR could not open arguments\n");
        exit(1);
    }
    
    std::vector<char> bytes;
    std::vector<char*> args;
    
    // Read the file bytewise (fseek does not work!)
    while( !feof(fp) )
    {
        char c;
        fread(&c, 1, 1, fp);
        bytes.push_back(c);
        // If we find a \0, cut off the string.
        if( c == 0 )
        {
            int len = bytes.size();
            char *c = new char[len+1];
            c[len] = 0;
            for( int i = 0; i < len; i++ )
                c[i] = bytes[i];
            bytes.clear();
            args.push_back(c);
            //printf("cmd: %s\n",c);
        }
    }
    *argc = args.size();
    *argv = new char*[args.size()];
    for( int i = 0; i < *argc; i++ )
        (*argv)[i] = args[i];
}
#elif defined(__APPLE__)
extern "C" int *_NSGetArgc(void);
extern "C" char ***_NSGetArgv(void);
void Mpize::read_cmdline( int* argc, char*** argv )
{
    *argc = *(_NSGetArgc());
    *argv = *(_NSGetArgv());
}
#elif defined(WIN32) || defined(_WIN32) || defined(__WIN32)
extern int __argc;
extern char ** __argv;
void Mpize::read_cmdline( int* argc, char*** argv )
{
    *argc = __argc;
    *argv = __argv;
}
#elif defined(__GNUC__) || defined(__INTEL_COMPILER) || defined(__ICC)
    #  include <features.h>

/**
 * This only works for GLIBC 2.4 upwards
 * as the GNU .init_array mechanism passes
 * argc + argv to the initialization functions
 * (the System V ABI does NOT!)
 */
    #if __GLIBC_PREREQ(2, 4)

int myargc = 0;
char **myargv = (char**)"";

void myinitfunc( int argc, char **argv, char **envp )
{
    myargc = argc;
    myargv = argv;
}

__attribute__((section(".init_array")))
static void* init_array_entry = &myinitfunc;
    #else
        #warning MPIZE: No mechanism to get argc/argv arguments for MPI_INIT!
    #endif
void Mpize::read_cmdline( int* argc, char*** argv )
{
    *argc = myargc;
    *argv = myargv;
}
#else
#warning MPIZE: No mechanism to get argc/argv arguments for MPI_INIT!
void Mpize::read_cmdline( int* argc, char*** argv )
{
    *argc = 0;
    *argv = (char**)"";
}
#endif
/*EOF*/
